package cz.fidentis.analyst.registration;

import cz.fidentis.analyst.Logger;
import cz.fidentis.analyst.canvas.Canvas;
import cz.fidentis.analyst.core.ControlPanelAction;
import cz.fidentis.analyst.face.HumanFace;
import cz.fidentis.analyst.face.events.HausdorffDistanceComputed;
import cz.fidentis.analyst.face.events.HumanFaceEvent;
import cz.fidentis.analyst.face.events.HumanFaceListener;
import cz.fidentis.analyst.face.events.MeshChangedEvent;
import cz.fidentis.analyst.feature.FeaturePoint;
import cz.fidentis.analyst.icp.IcpTransformation;
import cz.fidentis.analyst.icp.IcpTransformer;
import cz.fidentis.analyst.icp.NoUndersampling;
import cz.fidentis.analyst.icp.RandomStrategy;
import cz.fidentis.analyst.icp.UndersamplingStrategy;
import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import cz.fidentis.analyst.procrustes.ProcrustesAnalysis;
import cz.fidentis.analyst.scene.DrawableFace;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.util.List;
import java.util.zip.DataFormatException;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFormattedTextField;
import javax.swing.JTabbedPane;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * Action listener for the ICP and Procrustes registration.
 * <p>
 * Besides the UX logic, this object stores the parameters and results of 
 * manual, ICP, or Procrustes registration (alignment of human faces) 
 * <p/>
 * <p>
 * This object also serves as {@code HumanFaceListener}.
 * It means that it is invoked whenever one of the faces are changed and then can 
 * react to these changes.
 * <p/>
 * <p>
 * Changes made by this objects are announced to other listeners. 
 * Following events are triggered:
 * <ul>
 * <li>{@code MeshChangedEvent}</li>
 * <li></li>
 * </ul>
 * </p>
 *
 * @author Richard Pajersky
 * @author Radek Oslejsek
 * @author Daniel Schramm
 */
public class RegistrationAction extends ControlPanelAction implements HumanFaceListener  {

    /*
     * Attributes handling the state
     */
    private boolean scale = true;
    private int maxIterations = 10;
    private double error = 0.3;
    private UndersamplingStrategy undersampling = new RandomStrategy(200);
    private String strategy = RegistrationPanel.STRATEGY_POINT_TO_POINT;
    private boolean relativeDist = false;
    private boolean heatmapRender = false;
    private boolean procrustesScalingEnabled = false;


    /*
     * Coloring threshold and statistical values of feature point distances:
     */
    private double fpThreshold = 5.0;

    private final RegistrationPanel controlPanel;

    /**
     * Constructor.
     *
     * @param canvas OpenGL canvas
     * @param topControlPanel Top component for placing control panels
     */
    public RegistrationAction(Canvas canvas, JTabbedPane topControlPanel) {
        super(canvas, topControlPanel);
        this.controlPanel = new RegistrationPanel(this);

        // Place control panel to the topControlPanel
        topControlPanel.addTab(controlPanel.getName(), controlPanel.getIcon(), controlPanel);
        topControlPanel.addChangeListener(e -> {
            // If the registration panel is focused...
            if (((JTabbedPane) e.getSource()).getSelectedComponent() instanceof RegistrationPanel) {
                // ... display heatmap and feature points relevant to the registration
                //getCanvas().getScene().setDefaultColors();
                highlightCloseFeaturePoints();
                //getSecondaryDrawableFace().setRenderHeatmap(heatmapRender);
            }
        });
        topControlPanel.setSelectedComponent(controlPanel); // Focus registration panel
        //calculateHausdorffDistance();
        highlightCloseFeaturePoints();
        
        // Be informed about changes in faces perfomed by other GUI elements
        getPrimaryDrawableFace().getHumanFace().registerListener(this);
        getSecondaryDrawableFace().getHumanFace().registerListener(this);
    }

    @Override
    public void actionPerformed(ActionEvent ae) {
        double value;
        String action = ae.getActionCommand();

//        OutputWindow.print(ae.getActionCommand());
        switch (action) {
            case RegistrationPanel.ACTION_COMMAND_APPLY_ICP:
                applyICP();
                highlightCloseFeaturePoints();
                announceMeshChange(getSecondaryDrawableFace());
                break;
            case RegistrationPanel.ACTION_COMMAND_SHIFT_X:
                value = ((Number) (((JFormattedTextField) ae.getSource()).getValue())).doubleValue();
                getSecondaryDrawableFace().getTranslation().x = value;
                if (getSecondaryFeaturePoints() != null) {
                    getSecondaryFeaturePoints().getTranslation().x = value;
                }
                highlightCloseFeaturePoints();
                break;
            case RegistrationPanel.ACTION_COMMAND_SHIFT_Y:
                value = ((Number) (((JFormattedTextField) ae.getSource()).getValue())).doubleValue();
                getSecondaryDrawableFace().getTranslation().y = value;
                if (getSecondaryFeaturePoints() != null) {
                    getSecondaryFeaturePoints().getTranslation().y = value;
                }
                highlightCloseFeaturePoints();
                break;
            case RegistrationPanel.ACTION_COMMAND_SHIFT_Z:
                value = ((Number) (((JFormattedTextField) ae.getSource()).getValue())).doubleValue();
                getSecondaryDrawableFace().getTranslation().z = value;
                if (getSecondaryFeaturePoints() != null) {
                    getSecondaryFeaturePoints().getTranslation().z = value;
                }
                highlightCloseFeaturePoints();
                break;
            case RegistrationPanel.ACTION_COMMAND_ROTATE_X:
                value = ((Number) (((JFormattedTextField) ae.getSource()).getValue())).doubleValue();
                getSecondaryDrawableFace().getRotation().x = value;
                if (getSecondaryFeaturePoints() != null) {
                    getSecondaryFeaturePoints().getRotation().x = value;
                }
                highlightCloseFeaturePoints();
                break;
            case RegistrationPanel.ACTION_COMMAND_ROTATE_Y:
                value = ((Number) (((JFormattedTextField) ae.getSource()).getValue())).doubleValue();
                getSecondaryDrawableFace().getRotation().y = value;
                if (getSecondaryFeaturePoints() != null) {
                    getSecondaryFeaturePoints().getRotation().y = value;
                }
                highlightCloseFeaturePoints();
                break;
            case RegistrationPanel.ACTION_COMMAND_ROTATE_Z:
                value = ((Number) (((JFormattedTextField) ae.getSource()).getValue())).doubleValue();
                getSecondaryDrawableFace().getRotation().z = value;
                if (getSecondaryFeaturePoints() != null) {
                    getSecondaryFeaturePoints().getRotation().z = value;
                }
                highlightCloseFeaturePoints();
                break;
            case RegistrationPanel.ACTION_COMMAND_SCALE:
                value = ((Number) (((JFormattedTextField) ae.getSource()).getValue())).doubleValue();
                getSecondaryDrawableFace().getScale().x = value;
                getSecondaryDrawableFace().getScale().y = value;
                getSecondaryDrawableFace().getScale().z = value;
                if (getSecondaryFeaturePoints() != null) {
                    getSecondaryFeaturePoints().getScale().x = value;
                    getSecondaryFeaturePoints().getScale().y = value;
                    getSecondaryFeaturePoints().getScale().z = value;
                }
                highlightCloseFeaturePoints();
                break;
            case RegistrationPanel.ACTION_COMMAND_RESET_TRANSLATION:
                getSecondaryDrawableFace().setTranslation(new Vector3d(0, 0, 0));
                if (getSecondaryFeaturePoints() != null) {
                    getSecondaryFeaturePoints().setTranslation(new Vector3d(0, 0, 0));
                }
                highlightCloseFeaturePoints();
                break;
            case RegistrationPanel.ACTION_COMMAND_RESET_ROTATION:
                getSecondaryDrawableFace().setRotation(new Vector3d(0, 0, 0));
                if (getSecondaryFeaturePoints() != null) {
                    getSecondaryFeaturePoints().setRotation(new Vector3d(0, 0, 0));
                }
                highlightCloseFeaturePoints();
                break;
            case RegistrationPanel.ACTION_COMMAND_RESET_SCALE:
                getSecondaryDrawableFace().setScale(new Vector3d(0, 0, 0));
                if (getSecondaryFeaturePoints() != null) {
                    getSecondaryFeaturePoints().setScale(new Vector3d(0, 0, 0));
                }
                highlightCloseFeaturePoints();
                break;
            case RegistrationPanel.ACTION_COMMAND_APPLY_TRANSFORMATIONS:
                transformFace();
                getSecondaryDrawableFace().setTranslation(new Vector3d(0, 0, 0));
                getSecondaryDrawableFace().setRotation(new Vector3d(0, 0, 0));
                getSecondaryDrawableFace().setScale(new Vector3d(0, 0, 0));
                if (getSecondaryFeaturePoints() != null) {
                    getSecondaryFeaturePoints().setTranslation(new Vector3d(0, 0, 0));
                    getSecondaryFeaturePoints().setRotation(new Vector3d(0, 0, 0));
                    getSecondaryFeaturePoints().setScale(new Vector3d(0, 0, 0));
                }
                announceMeshChange(getSecondaryDrawableFace());
                break;
            case RegistrationPanel.ACTION_COMMAND_FP_CLOSENESS_THRESHOLD:
                fpThreshold = ((Number) (((JFormattedTextField) ae.getSource()).getValue())).doubleValue();
                highlightCloseFeaturePoints();
                break;
            case RegistrationPanel.ACTION_COMMAND_ICP_SCALE:
                this.scale = ((JCheckBox) ae.getSource()).isSelected();
                break;
            case RegistrationPanel.ACTION_COMMAND_ICP_MAX_ITERATIONS:
                maxIterations = ((Number) (((JFormattedTextField) ae.getSource()).getValue())).intValue();
                break;
            case RegistrationPanel.ACTION_COMMAND_ICP_ERROR:
                error = ((Number) (((JFormattedTextField) ae.getSource()).getValue())).doubleValue();
                break;
            case RegistrationPanel.ACTION_COMMAND_ICP_UNDERSAMPLING:
                String item = (String) ((JComboBox) ae.getSource()).getSelectedItem();
                switch (item) {
                    case "None":
                        this.undersampling = new NoUndersampling();
                        break;
                    case "Random 200":
                        this.undersampling = new RandomStrategy(200);
                        break;
                    default:
                        throw new UnsupportedOperationException(item);
                }
                break;
            case RegistrationPanel.ACTION_COMMAND_PROCRUSTES_SCALE:
                this.procrustesScalingEnabled = ((JCheckBox) ae.getSource()).isSelected();
                break;
            case RegistrationPanel.ACTION_COMMAND_PROCRUSTES_APPLY:
                applyProcrustes();
                highlightCloseFeaturePoints();
                announceMeshChange(getPrimaryDrawableFace());
                announceMeshChange(getSecondaryDrawableFace());
                break;
            default:
                // do nothing
        }

        renderScene();
    }
    
    @Override
    public void acceptEvent(HumanFaceEvent event) {
        if (event.getIssuer() == this) {
            return;
        }
        
        if (event instanceof HausdorffDistanceComputed) { // update stats
            HausdorffDistanceComputed hdEvent = (HausdorffDistanceComputed) event;
            controlPanel.updateHausdorffDistanceStats(
                    hdEvent.getHusdorffDistStats(),
                    hdEvent.getWeightedHusdorffDistStats()
            );
        }
    }
    
    /**
     * Calculates Procrustes analysis.
     * 
     * First it creates object of type ProcrusteAnalasys containing two face models converted
     * to ProcrustesAnalysisFaceModel objects which are required for next analysis step.
     * 
     * In the analysis step, faces are superimposed and rotated.
     * 
     * If the {@code procrustesScalingEnabled} attribute is set to true, then one of the faces is scaled as well.
     */
    protected void applyProcrustes() {
        Logger out = Logger.measureTime();

        HumanFace primaryFace = getCanvas().getPrimaryFace();
        HumanFace secondaryFace = getCanvas().getSecondaryFace();

        try {
            ProcrustesAnalysis procrustesAnalysisInit = new ProcrustesAnalysis(primaryFace, secondaryFace, this.procrustesScalingEnabled);
            procrustesAnalysisInit.analyze();
            primaryFace.removeKdTree(); // invalidate k-d tree, if exists
            secondaryFace.removeKdTree(); // invalidate k-d tree, if exists
            highlightCloseFeaturePoints();
        } catch (DataFormatException e) {
            Logger.print("Procrustes Analysis experienced exception");
        }

        out.printDuration("Procrustes for models with "
                + getPrimaryDrawableFace().getHumanFace().getMeshModel().getNumVertices()
                + "/"
                + getSecondaryDrawableFace().getHumanFace().getMeshModel().getNumVertices()
                + " vertices and "
                + getPrimaryDrawableFace().getHumanFace().getFeaturePoints().size()
                + " feature points"
        );
    }
    
    protected void applyICP() {
        Logger out = Logger.measureTime();

        IcpTransformer visitor = new IcpTransformer(getPrimaryDrawableFace().getModel(), maxIterations, scale, error, undersampling);
        getSecondaryDrawableFace().getModel().compute(visitor); // NOTE: the secondary face is physically transformed
        getSecondaryDrawableFace().getHumanFace().removeKdTree(); // invalidate k-d tree, if exists
        for (List<IcpTransformation> trList : visitor.getTransformations().values()) { // transform feature points
            for (IcpTransformation tr : trList) {
                for (int i = 0; i < getSecondaryFeaturePoints().getFeaturePoints().size(); i++) {
                    FeaturePoint fp = getSecondaryFeaturePoints().getFeaturePoints().get(i);
                    Point3d trPoint = tr.transformPoint(fp.getPosition(), scale);
                    getSecondaryFeaturePoints().getFeaturePoints().set(i, new FeaturePoint(trPoint.x, trPoint.y, trPoint.z, fp.getFeaturePointType()));
                }
            }
        }

        out.printDuration("ICP for models with "
                + getPrimaryDrawableFace().getHumanFace().getMeshModel().getNumVertices()
                + "/"
                + getSecondaryDrawableFace().getHumanFace().getMeshModel().getNumVertices()
                + " vertices"
        );
    }
    
    /**
     * Calculates feature points which are too far away and changes their color
     * to green otherwise set color to default
     */
    private void highlightCloseFeaturePoints() {
        if (getPrimaryDrawableFace() == null) { // scene not yet initiated
            return;
        }

        if (getPrimaryFeaturePoints() == null
                || getSecondaryFeaturePoints() == null
                || getPrimaryFeaturePoints().getFeaturePoints().size() != getSecondaryFeaturePoints().getFeaturePoints().size()) {
            return;
        }

        double fpMaxDist = Double.NEGATIVE_INFINITY;
        double fpMinDist = Double.POSITIVE_INFINITY;
        double distSum = 0.0;
        for (int i = 0; i < getPrimaryFeaturePoints().getFeaturePoints().size(); i++) {
            FeaturePoint primary = getPrimaryFeaturePoints().getFeaturePoints().get(i);
            FeaturePoint secondary = getSecondaryFeaturePoints().getFeaturePoints().get(i);
            Point3d transformed = new Point3d(secondary.getX(), secondary.getY(), secondary.getZ());
            transformPoint(transformed);
            double distance = Math.sqrt(
                Math.pow(transformed.x - primary.getX(), 2) + 
                Math.pow(transformed.y - primary.getY(), 2) + 
                Math.pow(transformed.z - primary.getZ(), 2));
            if (distance > fpThreshold) {
                getPrimaryFeaturePoints().resetColorToDefault(i);
                getSecondaryFeaturePoints().resetColorToDefault(i);
            } else {
                getPrimaryFeaturePoints().setColor(i, Color.GREEN);
                getSecondaryFeaturePoints().setColor(i, Color.GREEN);
            }
            fpMaxDist = Math.max(fpMaxDist, distance);
            fpMinDist = Math.min(fpMinDist, distance);
            distSum += distance;
        }
        double fpAvgDist = distSum / getPrimaryFeaturePoints().getFeaturePoints().size();
        // to do: show ACF dist
    }

    /**
     * Applies carried out transformations first on
     * {@link #getSecondaryDrawableFace} and then on
     * {@link #getSecondaryDrawableFace} feature points
     */
    private void transformFace() {
        for (MeshFacet transformedFacet : getSecondaryDrawableFace().getFacets()) {
            for (MeshPoint comparedPoint : transformedFacet.getVertices()) {
                transformPoint(comparedPoint.getPosition());
            }
        }
        for (int i = 0; i < getSecondaryFeaturePoints().getFeaturePoints().size(); i++) {
            FeaturePoint point = getSecondaryFeaturePoints().getFeaturePoints().get(i);
            Point3d transformed = new Point3d(point.getX(), point.getY(), point.getZ());
            transformPoint(transformed);
            point = new FeaturePoint(transformed.x, transformed.y, transformed.z, point.getFeaturePointType());
            getSecondaryFeaturePoints().getFeaturePoints().set(i, point);
        }
        getSecondaryDrawableFace().getHumanFace().removeKdTree(); // invalidate k-d tree, if exists
    }

    /**
     * Transforms point based on transformation info from
     * {@link #getSecondaryDrawableFace}
     *
     * @param point Point to transform
     */
    private void transformPoint(Point3d point) {
        if (point == null) {
            throw new IllegalArgumentException("point is null");
        }

        Point3d newPoint = new Point3d(0, 0, 0);
        double quotient;

        // rotate around X
        quotient = Math.toRadians(getSecondaryDrawableFace().getRotation().x);
        if (!Double.isNaN(quotient)) {
            double cos = Math.cos(quotient);
            double sin = Math.sin(quotient);
            newPoint.y = point.y * cos - point.z * sin;
            newPoint.z = point.z * cos + point.y * sin;
            point.y = newPoint.y;
            point.z = newPoint.z;
        }

        // rotate around Y
        quotient = Math.toRadians(getSecondaryDrawableFace().getRotation().y);
        if (!Double.isNaN(quotient)) {
            double cos = Math.cos(quotient);
            double sin = Math.sin(quotient);
            newPoint.x = point.x * cos + point.z * sin;
            newPoint.z = point.z * cos - point.x * sin;
            point.x = newPoint.x;
            point.z = newPoint.z;
        }

        // rotate around Z
        quotient = Math.toRadians(getSecondaryDrawableFace().getRotation().z);
        if (!Double.isNaN(quotient)) {
            double cos = Math.cos(quotient);
            double sin = Math.sin(quotient);
            newPoint.x = point.x * cos - point.y * sin;
            newPoint.y = point.y * cos + point.x * sin;
            point.x = newPoint.x;
            point.y = newPoint.y;
        }

        // translate
        point.x += getSecondaryDrawableFace().getTranslation().x;
        point.y += getSecondaryDrawableFace().getTranslation().y;
        point.z += getSecondaryDrawableFace().getTranslation().z;

        // scale
        point.x *= 1 + getSecondaryDrawableFace().getScale().x;
        point.y *= 1 + getSecondaryDrawableFace().getScale().y;
        point.z *= 1 + getSecondaryDrawableFace().getScale().z;
    }
    
    protected void announceMeshChange(DrawableFace face) {
        if (face != null) {
            face.getHumanFace().announceEvent(new MeshChangedEvent(face.getHumanFace(), face.getHumanFace().getShortName(), this));
        }
    }
}
